"""Module containing service classes and functions"""

import string
import random


def random_text(size):
    """Returns a random text of the specified size

    :param size: Size of the random string, must be greater than 0
    :type size int
    :return: Random string
    :rtype: str"""
    return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(size))



import unittest
from pythonping import utils


class UtilsTestCase(unittest.TestCase):
    """Tests for Utils class"""

    def test_random_text_generation(self):
        """Verifies that random_text generates text of correct size"""
        sizes = [0, 4, 1500, 1, 33, 44, 11]
        for size in sizes:
            self.assertEqual(
                len(utils.random_text(size)), size,
                'Unable to generate a random string of {0} characters'.format(size))



import socket
import select
import time


class Socket:
    DONT_FRAGMENT = (socket.SOL_IP, 10, 1)           # Option value for raw socket

    def __init__(self, destination, protocol, source=None, options=(), buffer_size=2048):
        """Creates a network socket to exchange messages

        :param destination: Destination IP address
        :type destination: str
        :param protocol: Name of the protocol to use
        :type protocol: str
        :param options: Options to set on the socket
        :type options: tuple
        :param source: Source IP to use - implemented in future releases
        :type source: Union[None, str]
        :param buffer_size: Size in bytes of the listening buffer for incoming packets (replies)
        :type buffer_size: int"""
        try:
            self.destination = socket.gethostbyname(destination)
        except socket.gaierror as e:
            raise RuntimeError('Cannot resolve address "' + destination + '", try verify your DNS or host file')
        self.protocol = socket.getprotobyname(protocol)
        self.buffer_size = buffer_size
        if source is not None:
            raise NotImplementedError('PythonPing currently does not support specification of source IP')
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_RAW, self.protocol)
        if options:
            self.socket.setsockopt(*options)

    def send(self, packet):
        """Sends a raw packet on the stream

        :param packet: The raw packet to send
        :type packet: bytes"""
        self.socket.sendto(packet, (self.destination, 0))

    def receive(self, timeout=2):
        """Listen for incoming packets until timeout

        :param timeout: Time after which stop listening
        :type timeout: Union[int, float]
        :return: The packet, the remote socket, and the time left before timeout
        :rtype: (bytes, tuple, float)"""
        time_left = timeout
        while time_left > 0:
            start_select = time.perf_counter()
            data_ready = select.select([self.socket], [], [], time_left)
            elapsed_in_select = time.perf_counter() - start_select
            time_left -= elapsed_in_select
            if not data_ready[0]:
                # Timeout
                return b'', '', time_left
            packet, source = self.socket.recvfrom(self.buffer_size)
            return packet, source, time_left

    def __del__(self):
        try:
            if hasattr(self, "socket") and self.socket:
                self.socket.close()
        except AttributeError:
            raise AttributeError("Attribute error because of failed socket init. Make sure you have the root privilege."
                                 " This error may also be caused from DNS resolution problems.")


import unittest
from pythonping.network import Socket


class UtilsTestCase(unittest.TestCase):
    """Tests for Socket class"""

    def test_raise_explicative_error_on_name_resolution_failure(self):
        """Test a runtime error is generated if the name cannot be resolved"""
        with self.assertRaises(RuntimeError):
            Socket('cant.resolve.this.address.localhost', 'ip')



import os
import socket
import struct
import select
import time


def checksum(data):
    """Creates the ICMP checksum as in RFC 1071

    :param data: Data to calculate the checksum ofs
    :type data: bytes
    :return: Calculated checksum
    :rtype: int

    Divides the data in 16-bits chunks, then make their 1's complement sum"""
    subtotal = 0
    for i in range(0, len(data)-1, 2):
        subtotal += ((data[i] << 8) + data[i+1])                # Sum 16 bits chunks together
    if len(data) % 2:                                           # If length is odd
        subtotal += (data[len(data)-1] << 8)                    # Sum the last byte plus one empty byte of padding
    while subtotal >> 16:                                       # Add carry on the right until fits in 16 bits
        subtotal = (subtotal & 0xFFFF) + (subtotal >> 16)
    check = ~subtotal                                           # Performs the one complement
    return ((check << 8) & 0xFF00) | ((check >> 8) & 0x00FF)    # Swap bytes


class ICMPType:
    """Represents an ICMP type, as combination of type and code

    ICMP Types should inherit from this class so that the code can identify them easily.
    This is a static class, not meant to be instantiated"""
    def __init__(self):
        raise TypeError('ICMPType may not be instantiated')


class Types(ICMPType):
    class EchoReply(ICMPType):
        type_id = 0
        ECHO_REPLY = (type_id, 0,)

    class DestinationUnreachable(ICMPType):
        type_id = 3
        NETWORK_UNREACHABLE = (type_id, 0,)
        HOST_UNREACHABLE = (type_id, 1,)
        PROTOCOL_UNREACHABLE = (type_id, 2,)
        PORT_UNREACHABLE = (type_id, 3,)
        FRAGMENTATION_REQUIRED = (type_id, 4,)
        SOURCE_ROUTE_FAILED = (type_id, 5,)
        NETWORK_UNKNOWN = (type_id, 6,)
        HOST_UNKNOWN = (type_id, 7,)
        SOURCE_HOST_ISOLATED = (type_id, 8,)
        NETWORK_ADMINISTRATIVELY_PROHIBITED = (type_id, 9,)
        HOST_ADMINISTRATIVELY_PROHIBITED = (type_id, 10,)
        NETWORK_UNREACHABLE_TOS = (type_id, 11,)
        HOST_UNREACHABLE_TOS = (type_id, 12,)
        COMMUNICATION_ADMINISTRATIVELY_PROHIBITED = (type_id, 13,)
        HOST_PRECEDENCE_VIOLATION = (type_id, 14,)
        PRECEDENCE_CUTOFF = (type_id, 15,)

    class SourceQuench(ICMPType):
        type_id = 4
        SOURCE_QUENCH = (type_id, 0,)
        
    class Redirect(ICMPType):
        type_id = 5
        FOR_NETWORK = (type_id, 0,)
        FOR_HOST = (type_id, 1,)
        FOR_TOS_AND_NETWORK = (type_id, 2,)
        FOR_TOS_AND_HOST = (type_id, 3,)
        
    class EchoRequest(ICMPType):
        type_id = 8
        ECHO_REQUEST = (type_id, 0,)
        
    class RouterAdvertisement(ICMPType):
        type_id = 9
        ROUTER_ADVERTISEMENT = (type_id, 0,)
        
    class RouterSolicitation(ICMPType):
        type_id = 10
        ROUTER_SOLICITATION = (type_id, 0)
        # Aliases
        ROUTER_DISCOVERY = ROUTER_SOLICITATION
        ROUTER_SELECTION = ROUTER_SOLICITATION
        
    class TimeExceeded(ICMPType):
        type_id = 11
        TTL_EXPIRED_IN_TRANSIT = (type_id, 0)
        FRAGMENT_REASSEMBLY_TIME_EXCEEDED = (type_id, 1)
        
    class BadIPHeader(ICMPType):
        type_id = 12
        POINTER_INDICATES_ERROR = (type_id, 0)
        MISSING_REQUIRED_OPTION = (type_id, 1)
        BAD_LENGTH = (type_id, 2)
        
    class Timestamp(ICMPType):
        type_id = 13
        TIMESTAMP = (type_id, 0)
        
    class TimestampReply(ICMPType):
        type_id = 14
        TIMESTAMP_REPLY = (type_id, 0)
        
    class InformationRequest(ICMPType):
        type_id = 15
        INFORMATION_REQUEST = (type_id, 0)
        
    class InformationReply(ICMPType):
        type_id = 16
        INFORMATION_REPLY = (type_id, 0)
        
    class AddressMaskRequest(ICMPType):
        type_id = 17
        ADDRESS_MASK_REQUEST = (type_id, 0)
        
    class AddressMaskReply(ICMPType):
        type_id = 18
        ADDRESS_MASK_REPLY = (type_id, 0)
        
    class Traceroute(ICMPType):
        type_id = 30
        INFORMATION_REQUEST = (type_id, 30)


class ICMP:
    LEN_TO_PAYLOAD = 41     # Ethernet, IP and ICMP header lengths combined

    def __init__(self, message_type=Types.EchoReply, payload=None, identifier=None, sequence_number=1):
        """Creates an ICMP packet

        :param message_type: Type of ICMP message to send
        :type message_type: Union[ICMPType, (int, int), int]
        :param payload: utf8 string or bytes payload
        :type payload: Union[str, bytes]
        :param identifier: ID of this ICMP packet
        :type identifier: int"""
        self.message_code = 0
        if issubclass(message_type, ICMPType):
            self.message_type = message_type.type_id
        elif isinstance(message_type, tuple):
            self.message_type = message_type[0]
            self.message_code = message_type[1]
        elif isinstance(message_type, int):
            self.message_type = message_type
        if payload is None:
            payload = bytes('1', 'utf8')
        elif isinstance(payload, str):
            payload = bytes(payload, 'utf8')
        self.payload = payload
        if identifier is None:
            identifier = os.getpid()
        self.id = identifier & 0xFFFF           # Prevent identifiers bigger than 16 bits
        self.sequence_number = sequence_number
        self.received_checksum = None

    @property
    def packet(self):
        """The raw packet with header, ready to be sent from a socket"""
        return self._header(check=self.expected_checksum) + self.payload

    def _header(self, check=0):
        """The raw ICMP header

        :param check: Checksum value
        :type check: int
        :return: The packed header
        :rtype: bytes"""
        # TODO implement sequence number
        return struct.pack("bbHHh",
                           self.message_type,
                           self.message_code,
                           check,
                           self.id,
                           self.sequence_number)

    @property
    def is_valid(self):
        """True if the received checksum is valid, otherwise False"""
        if self.received_checksum is None:
            return True
        return self.expected_checksum == self.received_checksum

    @property
    def expected_checksum(self):
        """The checksum expected for this packet, calculated with checksum field set to 0"""
        return checksum(self._header() + self.payload)

    @property
    def header_length(self):
        """Length of the ICMP header"""
        return len(self._header())

    @staticmethod
    def generate_from_raw(raw):
        """Creates a new ICMP representation from the raw bytes

        :param raw: The raw packet including payload
        :type raw: bytes
        :return: An ICMP instance representing the packet
        :rtype: ICMP"""
        packet = ICMP()
        packet.unpack(raw)
        return packet

    def unpack(self, raw):
        """Unpacks a raw packet and stores it in this object

        :param raw: The raw packet, including payload
        :type raw: bytes"""
        self.message_type, \
            self.message_code, \
            self.received_checksum, \
            self.id, \
            sequence = struct.unpack("bbHHh", raw[20:28])
        self.payload = raw[28:]



"""Module generating ICMP payloads (with no header)"""


class PayloadProvider:
    def __init__(self):
        raise NotImplementedError('Cannot create instances of PayloadProvider')

    def __iter__(self):
        raise NotImplementedError()

    def __next__(self):
        raise NotImplementedError()


class List(PayloadProvider):
    def __init__(self, payload_list):
        """Creates a provider of payloads from an existing list of payloads

        :param payload_list: An existing list of payloads
        :type payload_list: list"""
        self._payloads = payload_list
        self._counter = 0

    def __iter__(self):
        self._counter = 0
        return self

    def __next__(self):
        if self._counter < len(self._payloads):
            ret = self._payloads[self._counter]
            self._counter += 1
            return ret
        raise StopIteration


class Repeat(PayloadProvider):
    def __init__(self, pattern, count):
        """Creates a provider of many identical payloads

        :param pattern: The existing payload
        :type pattern: Union[str, bytes]
        :param count: How many payloads to generate
        :type count: int"""
        self.pattern = pattern
        self.count = count
        self._counter = 0

    def __iter__(self):
        self._counter = 0
        return self

    def __next__(self):
        if self._counter < self.count:
            self._counter += 1
            return self.pattern
        raise StopIteration


class Sweep(PayloadProvider):
    def __init__(self, pattern, start_size, end_size):
        """Creates a provider of payloads of increasing size

        :param pattern: The existing payload, may be cut or replicated to fit the size
        :type pattern: Union[str, bytes]
        :param start_size: The first payload size to start with, included
        :type start_size: int
        :param end_size: The payload size to end with, included
        :type end_size: int"""
        if start_size > end_size:
            raise ValueError('end_size must be greater or equal than start_size')
        if len(pattern) == 0:
            raise ValueError('pattern cannot be empty')
        self.pattern = pattern
        self.start_size = start_size
        self.end_size = end_size
        # Extend the length of the pattern if needed
        while not len(self.pattern) >= end_size:
            self.pattern += pattern
        self._current_size = self.start_size

    def __iter__(self):
        self._current_size = self.start_size
        return self

    def __next__(self):
        if self._current_size <= self.end_size:
            ret = self.pattern[0:self._current_size]
            self._current_size += 1
            return ret
        raise StopIteration



import unittest
from pythonping import icmp


class ICMPTestCase(unittest.TestCase):
    """Tests for the ICMP class"""

    def test_checksum(self):
        """Verifies that checksum calculation is correct"""
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\x00*\x01\x009TPM'),
            13421,
            "Checksum validation failed (even length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\x01*\x01\x001PJ2'),
            21370,
            "Checksum validation failed (even length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\x02*\x01\x006K3J'),
            16523,
            "Checksum validation failed (even length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\x03*\x01\x00COSA'),
            17757,
            "Checksum validation failed (even length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\xb4$\x01\x0012344'),
            29866,
            "Checksum validation failed (odd length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00$F\x01\x00fag'),
            22533,
            "Checksum validation failed (odd length)"
        )
        self.assertEqual(
            icmp.checksum(b'\x08\x00\x00\x00\x8cH\x01\x00random text goes here'),
            47464,
            "Checksum validation failed (odd length)"
        )

    def test_pack(self):
        """Verifies that creates the correct pack"""
        self.assertEqual(
            icmp.ICMP(icmp.Types.EchoReply, payload='banana', identifier=19700).packet,
            b'\x00\x00\xcb\x8e\xf4L\x01\x00banana',
            "Fail to pack ICMP structure to packet"
        )
        self.assertEqual(
            icmp.ICMP(icmp.Types.EchoReply, payload='random text goes here', identifier=12436).packet,
            b'\x00\x00h\xd1\x940\x01\x00random text goes here',
            "Fail to pack ICMP structure to packet"
        )
        self.assertEqual(
            icmp.ICMP(icmp.Types.EchoRequest, payload='random text goes here', identifier=18676).packet,
            b'\x08\x00\x00\xb9\xf4H\x01\x00random text goes here',
            "Fail to unpack ICMP structure to packet"
        )

    def test_unpack(self):
        """Verifies that reads data correctly from a packet"""
        ip_header_offset = b''.join([b'0' for _ in range(20)])

        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x00\x00\xbe\xdb\x01\x00\x01\x00banana')
        self.assertEqual(packet.message_type, 0, 'Failed to extract message type')
        self.assertEqual(packet.message_code, 0, 'Failed to extract message code')
        self.assertEqual(packet.id, 1, 'Failed to extract id')
        self.assertEqual(packet.payload, b'banana', 'Failed to extract payload')

        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x00\x00Q\x9d\x10\x00\x01\x00random text goes here')
        self.assertEqual(packet.message_type, 0, 'Failed to extract message type')
        self.assertEqual(packet.message_code, 0, 'Failed to extract message code')
        self.assertEqual(packet.id, 16, 'Failed to extract id')
        self.assertEqual(packet.payload, b'random text goes here', 'Failed to extract payload')

        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x08\x00I\x9d\x10\x00\x01\x00random text goes here')
        self.assertEqual(packet.message_type, 8, 'Failed to extract message type')
        self.assertEqual(packet.message_code, 0, 'Failed to extract message code')
        self.assertEqual(packet.id, 16, 'Failed to extract id')
        self.assertEqual(packet.payload, b'random text goes here', 'Failed to extract payload')

    def test_is_valid(self):
        """Verifies that understands if receives a packet with valid or invalid checksum"""
        ip_header_offset = b''.join([b'0' for _ in range(20)])
        # Following two packets have a good checksum
        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x00\x00\xbe\xdb\x01\x00\x01\x00banana')
        self.assertEqual(packet.received_checksum, packet.expected_checksum, 'Checksum validation failed')
        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x08\x00@\xca\xb47\x01\x00random text goes here')
        self.assertEqual(packet.received_checksum, packet.expected_checksum, 'Checksum validation failed')
        # Packet 'does' instead of 'goes'
        packet = icmp.ICMP.generate_from_raw(ip_header_offset + b'\x08\x00\x9c\xd0X1\x01\x00random text does here')
        self.assertNotEqual(packet.received_checksum, packet.expected_checksum, 'Checksum validation failed')

    def test_checksum_creation(self):
        """Verifies it generates the correct checksum, given packet data"""
        packet = icmp.ICMP(icmp.Types.EchoRequest, payload='random text goes here', identifier=16)
        self.assertEqual(packet.expected_checksum, 485, 'Checksum creation failed')
        packet = icmp.ICMP(icmp.Types.EchoReply, payload='foobar', identifier=11)
        self.assertEqual(packet.expected_checksum, 48060, 'Checksum creation failed')

    def test_blank_header_creation(self):
        """Verifies it generates the correct header (no checksum) given packet data"""
        packet = icmp.ICMP(icmp.Types.EchoRequest, payload='random text goes here', identifier=16)
        self.assertEqual(packet._header(), b'\x08\x00\x00\x00\x10\x00\x01\x00',
                         'Blank header creation failed (without checksum)')
        packet = icmp.ICMP(icmp.Types.EchoReply, payload='foo', identifier=11)
        self.assertEqual(packet._header(), b'\x00\x00\x00\x00\x0b\x00\x01\x00',
                         'Blank header creation failed (without checksum)')



import unittest
from pythonping import payload_provider


class PayloadProviderTestCase(unittest.TestCase):
    """Tests for PayloadProvider class"""

    def test_list(self):
        """Verifies that a list provider generates the correct payloads"""
        payloads = [
            b'payload A',
            b'second payload'
            b'payload 3+'
        ]
        res = []
        provider = payload_provider.List(payloads)
        for payload in provider:
            res.append(payload)
        for num, payload in enumerate(payloads):
            self.assertEqual(res[num], payload, 'Payload not expected in position {0}'.format(num))

    def test_repeat(self):
        """Verifies that a repeat provider generates the correct payloads"""
        pattern = b'this is a pattern'
        count = 5
        provider = payload_provider.Repeat(pattern, count)
        for payload in provider:
            self.assertEqual(payload, pattern, 'Payload does not reflect the pattern')
            count -= 1
        self.assertEqual(count, 0, 'Generated a wrong number of payloads')

    def sweep_tester(self, pattern, start, end):
        """Runs the creation of a sweep provider and performs some basics tests on it"""
        provider = payload_provider.Sweep(pattern, start, end)
        payloads_generated = 0
        for payload in provider:
            self.assertEqual(len(payload), start, 'Generated a payload with a wrong size')
            start += 1
            payloads_generated += 1
        self.assertEqual(start, end+1, 'Generated a wrong number of payloads')
        return payloads_generated

    def test_sweep_normal(self):
        """Verifies that a sweep provider generates the correct payloads"""
        self.sweep_tester(b'abc', 10, 20)

    def test_sweep_one(self):
        """Verifies that a sweep provider generates one correct payload if start and end sizes are equal"""
        self.assertEqual(
            self.sweep_tester(b'frog', 40, 40),
            1,
            'Unable to generate exactly 1 payload during a sweep')

    def test_sweep_reversed(self):
        """Verifies that it is not possible to generate a payload with start size greater than end size"""
        with self.assertRaises(ValueError):
            payload_provider.Sweep(b'123', 100, 45)

    def test_sweep_no_pattern(self):
        """Verifies that it is not possible to generate a payload with an empty pattern"""
        with self.assertRaises(ValueError):
            payload_provider.Sweep(b'', 1, 10)



import collections
import unittest
from pythonping import executor
from pythonping import icmp


class SuccessfulResponseMock(executor.Response):
    """Mock a successful response to a ping"""
    @property
    def success(self):
        return True


class FailingResponseMock(executor.Response):
    """Mock a failed response to a ping"""
    @property
    def success(self):
        return False


class ExecutorUtilsTestCase(unittest.TestCase):
    """Tests for standalone functions in pythonping.executor"""

    def test_represent_seconds_in_ms(self):
        """Verifies conversion from seconds to milliseconds works correctly"""
        self.assertEqual(executor.represent_seconds_in_ms(4), 4000, 'Failed to convert seconds to milliseconds')
        self.assertEqual(executor.represent_seconds_in_ms(0), 0, 'Failed to convert seconds to milliseconds')
        self.assertEqual(executor.represent_seconds_in_ms(0.001), 1, 'Failed to convert seconds to milliseconds')
        self.assertEqual(executor.represent_seconds_in_ms(0.0001), 0.1, 'Failed to convert seconds to milliseconds')
        self.assertEqual(executor.represent_seconds_in_ms(0.00001), 0.01, 'Failed to convert seconds to milliseconds')


class ResponseTestCase(unittest.TestCase):
    """Tests to verify that a Response object renders information correctly"""

    @staticmethod
    def craft_response_of_type(response_type):
        """Generates an executor.Response from an icmp.Types

        :param response_type: Type of response
        :type response_type: Union[icmp.Type, tuple]
        :return: The crafted response
        :rtype: executor.Response"""
        return executor.Response(executor.Message('', icmp.ICMP(response_type), '127.0.0.1'), 0.1)

    def test_success(self):
        """Verifies the if the Response can indicate a success to a request correctly"""
        self.assertTrue(self.craft_response_of_type(icmp.Types.EchoReply).success,
                        'Unable to validate a successful response')
        self.assertFalse(self.craft_response_of_type(icmp.Types.DestinationUnreachable).success,
                         'Unable to validate Destination Unreachable')
        self.assertFalse(self.craft_response_of_type(icmp.Types.BadIPHeader).success,
                         'Unable to validate Bad IP Header')
        self.assertFalse(executor.Response(None, 0.1).success, 'Unable to validate timeout (no payload)')

    def test_error_message(self):
        """Verifies error messages are presented correctly"""
        self.assertEqual(self.craft_response_of_type(icmp.Types.EchoReply).error_message, None,
                         'Generated error message when response was correct')
        self.assertEqual(self.craft_response_of_type(icmp.Types.DestinationUnreachable).error_message,
                         'Network Unreachable',
                         'Unable to classify a Network Unreachable error correctly')
        self.assertEqual(executor.Response(None, 0.1).error_message, 'No response',
                         'Unable to generate correct message when response is not received')
        pass

    def time_elapsed(self):
        """Verifies the time elapsed is presented correctly"""
        self.assertEqual(executor.Response(None, 1).time_elapsed_ms, 1000, 'Bad ms representation for 1 second')
        self.assertEqual(executor.Response(None, 0.001).time_elapsed_ms, 1, 'Bad ms representation for 1 ms')
        self.assertEqual(executor.Response(None, 0.01).time_elapsed_ms, 10, 'Bad ms representation for 10 ms')


class ResponseListTestCase(unittest.TestCase):
    """Tests for ResponseList"""

    @staticmethod
    def responses_from_times(times):
        """Generates a list of empty responses from a list of time elapsed

        :param times: List of time elapsed for each response
        :type times: list
        :return: List of responses
        :rtype: executor.ResponseList"""
        return executor.ResponseList([executor.Response(None, _) for _ in times])

    def test_rtt_min_ms(self):
        """Verifies the minimum RTT is found correctly"""
        self.assertEqual(
            self.responses_from_times([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).rtt_min,
            0,
            'Unable to identify minimum RTT of 0'
        )
        self.assertEqual(
            self.responses_from_times([38, 11, 93, 100, 38, 11, 0.1]).rtt_min,
            0.1,
            'Unable to identify minimum RTT of 0.1'
        )
        self.assertEqual(
            self.responses_from_times([10, 10, 10, 10]).rtt_min,
            10,
            'Unable to identify minimum RTT of 10 on a series of only 10s'
        )

    def test_rtt_max_ms(self):
        """Verifies the maximum RTT is found correctly"""
        self.assertEqual(
            self.responses_from_times([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).rtt_max,
            9,
            'Unable to identify maximum RTT of 9'
        )
        self.assertEqual(
            self.responses_from_times([38, 11, 93, 100, 38, 11, 0.1]).rtt_max,
            100,
            'Unable to identify maximum RTT of 100'
        )
        self.assertEqual(
            self.responses_from_times([10, 10, 10, 10]).rtt_max,
            10,
            'Unable to identify maximum RTT of 10 on a series of only 10s'
        )

    def test_rtt_avg_ms(self):
        """Verifies the average RTT is found correctly"""
        self.assertEqual(
            self.responses_from_times([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).rtt_avg,
            4.5,
            'Unable to identify average RTT of 4.5'
        )
        self.assertEqual(
            self.responses_from_times([38, 11, 93, 100, 38, 11, 0.1]).rtt_avg,
            41.58571428571429,
            'Unable to identify average RTT of 41.58571428571429'
        )
        self.assertEqual(
            self.responses_from_times([10, 10, 10, 10]).rtt_avg,
            10,
            'Unable to identify average RTT of 10 on a series of only 10s'
        )

    def test_len(self):
        """Verifies the length is returned correctly"""
        self.assertEqual(
            len(self.responses_from_times(list(range(10)))),
            10,
            'Unable identify the length of 10'
        )
        self.assertEqual(
            len(self.responses_from_times(list(range(0)))),
            0,
            'Unable identify the length of 0'
        )
        self.assertEqual(
            len(self.responses_from_times(list(range(23)))),
            23,
            'Unable identify the length of 23'
        )

    def test_iterable(self):
        """Verifies it is iterable"""
        self.assertTrue(
            isinstance(self.responses_from_times([0, 1, 2, 3]), collections.abc.Iterable),
            'Unable to iterate over ResponseList object'
        )

    def test_success_all_success(self):
        """Verify success is calculated correctly if all responses are successful"""
        rs = executor.ResponseList([
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1)
        ])
        self.assertTrue(
            rs.success(executor.SuccessOn.One),
            'Unable to calculate success on one correctly with all responses successful'
        )
        self.assertTrue(
            rs.success(executor.SuccessOn.Most),
            'Unable to calculate success on most with all responses successful'
        )
        self.assertTrue(
            rs.success(executor.SuccessOn.All),
            'Unable to calculate success on all with all responses successful'
        )

    def test_success_one_success(self):
        """Verify success is calculated correctly if one response is successful"""
        rs = executor.ResponseList([
            SuccessfulResponseMock(None, 1),
            FailingResponseMock(None, 1),
            FailingResponseMock(None, 1),
            FailingResponseMock(None, 1)
        ])
        self.assertTrue(
            rs.success(executor.SuccessOn.One),
            'Unable to calculate success on one correctly with one response successful'
        )
        self.assertFalse(
            rs.success(executor.SuccessOn.Most),
            'Unable to calculate success on most with one response successful'
        )
        self.assertFalse(
            rs.success(executor.SuccessOn.All),
            'Unable to calculate success on all with one response successful'
        )

    def test_success_most_success(self):
        """Verify success is calculated correctly if most responses are successful"""
        rs = executor.ResponseList([
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1),
            FailingResponseMock(None, 1)
        ])
        self.assertTrue(
            rs.success(executor.SuccessOn.One),
            'Unable to calculate success on one correctly with most responses successful'
        )
        self.assertTrue(
            rs.success(executor.SuccessOn.Most),
            'Unable to calculate success on most with most responses successful'
        )
        self.assertFalse(
            rs.success(executor.SuccessOn.All),
            'Unable to calculate success on all with most responses successful'
        )

    def test_success_half_success(self):
        """Verify success is calculated correctly if half responses are successful"""
        rs = executor.ResponseList([
            SuccessfulResponseMock(None, 1),
            SuccessfulResponseMock(None, 1),
            FailingResponseMock(None, 1),
            FailingResponseMock(None, 1)
        ])
        self.assertTrue(
            rs.success(executor.SuccessOn.One),
            'Unable to calculate success on one correctly with half responses successful'
        )
        self.assertFalse(
            rs.success(executor.SuccessOn.Most),
            'Unable to calculate success on most with half responses successful'
        )
        self.assertFalse(
            rs.success(executor.SuccessOn.All),
            'Unable to calculate success on all with half responses successful'
        )


class CommunicatorTestCase(unittest.TestCase):
    """Tests for Communicator"""

    def test_increase_seq(self):
        """Verifies Communicator can increase sequence number correctly"""
        self.assertEqual(executor.Communicator.increase_seq(1), 2, 'Increasing sequence number 1 did not return 2')
        self.assertEqual(executor.Communicator.increase_seq(100), 101, 'Increasing sequence number 1 did not return 2')
        self.assertEqual(executor.Communicator.increase_seq(0xFFFF), 1,
                         'Not returned to 1 when exceeding sequence number maximum length')
        self.assertEqual(executor.Communicator.increase_seq(0xFFFE), 0xFFFF,
                         'Increasing sequence number 0xFFFE did not return 0xFFFF')



"""Module that actually performs the ping, sending and receiving packets"""

import os
import sys
from . import icmp
from . import network

# Python 3.5 compatibility
if sys.version_info[1] == 5:
    from enum import IntEnum, Enum

    class AutoNumber(Enum):
        def __new__(cls):
            value = len(cls.__members__) + 1
            obj = object.__new__(cls)
            obj._value = value
            return obj

    class SuccessOn(AutoNumber):
        One = ()
        Most = ()
        All = ()
else:
    from enum import IntEnum, auto

    class SuccessOn(IntEnum):
        One = auto()
        Most = auto()
        All = auto()


class Message:
    """Represents an ICMP message with destination socket"""
    def __init__(self, target, packet, source):
        """Creates a message that may be sent, or used to represent a response

        :param target: Target IP or hostname of the message
        :type target: str
        :param packet: ICMP packet composing the message
        :type packet: icmp.ICMP
        :param source: Source IP or hostname of the message
        :type source: str"""
        self.target = target
        self.packet = packet
        self.source = source

    def send(self, source_socket):
        """Places the message on a socket

        :param source_socket: The socket to place the message on
        :type source_socket: network.Socket"""
        source_socket.send(self.packet.packet)


def represent_seconds_in_ms(seconds):
    """Converts seconds into human-readable milliseconds with 2 digits decimal precision

    :param seconds: Seconds to convert
    :type seconds: Union[int, float]
    :return: The same time expressed in milliseconds, with 2 digits of decimal precision
    :rtype: float"""
    return round(seconds * 1000, 2)


class Response:
    """Represents a response to an ICMP message, with metadata like timing"""
    def __init__(self, message, time_elapsed):
        """Creates a representation of ICMP message received in response

        :param message: The message received
        :type message: Union[None, Message]
        :param time_elapsed: Time elapsed since the original request was sent, in seconds
        :type time_elapsed: float"""
        self.message = message
        self.time_elapsed = time_elapsed

    @property
    def success(self):
        return self.error_message is None

    @property
    def error_message(self):
        if self.message is None:
            return 'No response'
        else:
            if self.message.packet.message_type == 0 and self.message.packet.message_code == 0:
                # Echo Reply, response OK - no error
                return None
            elif self.message.packet.message_type == 3:
                # Destination unreachable, returning more details based on message code
                unreachable_messages = [
                    'Network Unreachable',
                    'Host Unreachable',
                    'Protocol Unreachable',
                    'Port Unreachable',
                    'Fragmentation Required',
                    'Source Route Failed',
                    'Network Unknown',
                    'Host Unknown',
                    'Source Host Isolated',
                    'Communication with Destination Network is Administratively Prohibited',
                    'Communication with Destination Host is Administratively Prohibited',
                    'Network Unreachable for ToS',
                    'Host Unreachable for ToS',
                    'Communication Administratively Prohibited',
                    'Host Precedence Violation',
                    'Precedence Cutoff in Effect'
                ]
                try:
                    return unreachable_messages[self.message.packet.message_code]
                except IndexError:
                    # Should never generate IndexError, this serves as additional protection
                    return 'Unreachable'
        # Error was not identified
        return 'Network Error'

    @property
    def time_elapsed_ms(self):
        return represent_seconds_in_ms(self.time_elapsed)

    def __repr__(self):
        if self.message is None:
            return 'Request timed out'
        elif self.success:
            return 'Reply from {0}, {1} bytes in {2}ms'.format(self.message.source,
                                                               len(self.message.packet.packet),
                                                               self.time_elapsed_ms)
        else:
            # Not successful, but with some code (e.g. destination unreachable)
            return '{0} from {1} in {2}ms'.format(self.error_message, self.message.source, self.time_elapsed_ms)


class ResponseList:
    """Represents a series of ICMP responses"""
    def __init__(self, initial_set=[], verbose=False, output=sys.stdout):
        """Creates a ResponseList with initial data if available

        :param initial_set: Already existing responses
        :type initial_set: list
        :param verbose: Flag to enable verbose mode, defaults to False
        :type verbose: bool
        :param output: File where to write verbose output, defaults to stdout
        :type output: file"""
        self._responses = []
        self.clear()
        self.verbose = verbose
        self.output = output
        self.rtt_avg = 0
        self.rtt_min = 0
        self.rtt_max = 0
        for response in initial_set:
            self.append(response)

    def success(self, option=SuccessOn.One):
        """Check success state of the request.

        :param option: Sets a threshold for success sign. ( 1 - SuccessOn.One, 2 - SuccessOn.Most, 3 - SuccessOn.All )
        :type option: int
        :return: Whether this set of responses is successful
        :rtype: bool
        """
        result = False
        success_list = [resp.success for resp in self._responses]
        if option == SuccessOn.One:
            result = True in success_list
        elif option == SuccessOn.Most:
            result = success_list.count(True) / len(success_list) > 0.5
        elif option == SuccessOn.All:
            result = False not in success_list
        return result

    @property
    def rtt_min_ms(self):
        return represent_seconds_in_ms(self.rtt_min)

    @property
    def rtt_max_ms(self):
        return represent_seconds_in_ms(self.rtt_max)

    @property
    def rtt_avg_ms(self):
        return represent_seconds_in_ms(self.rtt_avg)

    def clear(self):
        self._responses = []

    def append(self, value):
        self._responses.append(value)
        if len(self) == 1:
            self.rtt_avg = value.time_elapsed
            self.rtt_max = value.time_elapsed
            self.rtt_min = value.time_elapsed
        else:
            # Calculate the total of time, add the new value and divide for the new number
            self.rtt_avg = ((self.rtt_avg * (len(self)-1)) + value.time_elapsed) / len(self)
            if value.time_elapsed > self.rtt_max:
                self.rtt_max = value.time_elapsed
            if value.time_elapsed < self.rtt_min:
                self.rtt_min = value.time_elapsed
        if self.verbose:
            print(value, file=self.output)

    def __len__(self):
        return len(self._responses)

    def __repr__(self):
        ret = ''
        for response in self._responses:
            ret += '{0}\r\n'.format(response)
        ret += '\r\n'
        ret += 'Round Trip Times min/avg/max is {0}/{1}/{2} ms'.format(self.rtt_min_ms, self.rtt_avg_ms, self.rtt_max_ms)
        return ret

    def __iter__(self):
        for response in self._responses:
            yield response


class Communicator:
    """Instance actually communicating over the network, sending messages and handling responses"""
    def __init__(self, target, payload_provider, timeout, socket_options=(), seed_id=None,
                 verbose=False, output=sys.stdout):
        """Creates an instance that can handle communication with the target device

        :param target: IP or hostname of the remote device
        :type target: str
        :param payload_provider: An iterable list of payloads to send
        :type payload_provider: PayloadProvider
        :param timeout: Timeout that will apply to all ping messages, in seconds
        :type timeout: int
        :param socket_options: Options to specify for the network.Socket
        :type socket_options: tuple
        :param seed_id: The first ICMP packet ID to use
        :type seed_id: Union[None, int]
        :param verbose: Flag to enable verbose mode, defaults to False
        :type verbose: bool
        :param output: File where to write verbose output, defaults to stdout
        :type output: file"""
        self.socket = network.Socket(target, 'icmp', source=None, options=socket_options)
        self.provider = payload_provider
        self.timeout = timeout
        self.responses = ResponseList(verbose=verbose, output=output)
        self.seed_id = seed_id
        if self.seed_id is None:
            self.seed_id = os.getpid() & 0xFFFF

    def __del__(self):
        pass

    def send_ping(self, packet_id, sequence_number, payload):
        """Sends one ICMP Echo Request on the socket

        :param packet_id: The ID to use for the packet
        :type packet_id: int
        :param sequence_number: The sequence number to use for the packet
        :type sequence_number: int
        :param payload: The payload of the ICMP message
        :type payload: Union[str, bytes]
        :rtype: bytes"""
        i = icmp.ICMP(
            icmp.Types.EchoRequest,
            payload=payload,
            identifier=packet_id, sequence_number=sequence_number)
        self.socket.send(i.packet)
        return i.payload

    def listen_for(self, packet_id, timeout, payload_pattern=None):
        """Listens for a packet of a given id for a given timeout

        :param packet_id: The ID of the packet to listen for, the same for request and response
        :type packet_id: int
        :param timeout: How long to listen for the specified packet, in seconds
        :type timeout: float
        :param payload_pattern: Payload reply pattern to match to request, if set to None, match by ID only
        :type payload_pattern: Union[None, bytes]
        :return: The response to the request with the specified packet_id
        :rtype: Response"""
        time_left = timeout
        response = icmp.ICMP()
        while time_left > 0:
            # Keep listening until a packet arrives
            raw_packet, source_socket, time_left = self.socket.receive(time_left)
            # If we actually received something
            if raw_packet != b'':
                response.unpack(raw_packet)

                # Ensure we have not unpacked the packet we sent (RHEL will also listen to outgoing packets)
                if response.id == packet_id and response.message_type != icmp.Types.EchoRequest.type_id:
                    if payload_pattern is None:
                        # To allow Windows-like behaviour (no payload inspection, but only match packet identifiers),
                        # simply allow for it to be an always true in the legacy usage case
                        payload_matched = True
                    else:
                        payload_matched = (payload_pattern == response.payload)

                    if payload_matched:
                        return Response(Message('', response, source_socket[0]), timeout - time_left)
        return Response(None, timeout)

    @staticmethod
    def increase_seq(sequence_number):
        """Increases an ICMP sequence number and reset if it gets bigger than 2 bytes

        :param sequence_number: The sequence number to increase
        :type sequence_number: int
        :return: The increased sequence number of 1, in case an increase was not possible
        :rtype: int"""
        sequence_number += 1
        if sequence_number > 0xFFFF:
            sequence_number = 1
        return sequence_number

    def run(self, match_payloads=False):
        """Performs all the pings and stores the responses

        :param match_payloads: optional to set to True to make sure requests and replies have equivalent payloads
        :type match_payloads: bool"""
        self.responses.clear()
        identifier = self.seed_id
        seq = 1
        for payload in self.provider:
            payload_bytes_sent = self.send_ping(identifier, seq, payload)
            if not match_payloads:
                self.responses.append(self.listen_for(identifier, self.timeout))
            else:
                self.responses.append(self.listen_for(identifier, self.timeout, payload_bytes_sent))

            seq = self.increase_seq(seq)



from setuptools import setup

with open('README.md', 'r') as file:
    long_description = file.read()

setup(name='pythonping',
      version='1.0.12',
      description='A simple way to ping in Python',
      url='https://github.com/alessandromaggio/pythonping',
      author='Alessandro Maggio',
      author_email='me@alessandromaggio.com',
      license='MIT',
      packages=['pythonping'],
      keywords=['ping', 'icmp', 'network'],
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: System Administrators',
            'Natural Language :: English'
      ],
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False)



import sys
from . import network, executor, payload_provider
from .utils import random_text
from random import randint
from typing import List


# this needs to be available across all thread usages
SEED_IDs: List[int] = []


def ping(target,
         timeout=2,
         count=4,
         size=1,
         payload=None,
         sweep_start=None,
         sweep_end=None,
         df=False,
         verbose=False,
         out=sys.stdout,
         match=False):
    """Pings a remote host and handles the responses

    :param target: The remote hostname or IP address to ping
    :type target: str
    :param timeout: How long before considering each non-arrived reply permanently lost
    :type timeout: int
    :param count: How many times to attempt the ping
    :type count: int
    :param size: Size of the entire packet to send
    :type size: int
    :param payload: Payload content, leave None if size is set to use random text
    :type payload: Union[str, bytes]
    :param sweep_start: If size is not set, initial size in a sweep of sizes
    :type sweep_start: int
    :param sweep_end: If size is not set, final size in a sweep of sizes
    :type sweep_end: int
    :param df: Don't Fragment flag value for IP Header
    :type df: bool
    :param verbose: Print output while performing operations
    :type verbose: bool
    :param out: Stream to which redirect the verbose output
    :type out: stream
    :param match: Do payload matching between request and reply (default behaviour follows that of Windows which is
    by packet identifier only, Linux behaviour counts a non equivalent payload in reply as fail, such as when pinging
    8.8.8.8 with 1000 bytes and reply is truncated to only the first 74 of request payload with packet identifiers
    the same in request and reply)
    :type match: bool
    :return: List with the result of each ping
    :rtype: executor.ResponseList"""
    provider = payload_provider.Repeat(b'', 0)
    if size and size > 0:
        if not payload:
            payload = random_text(size)
        provider = payload_provider.Repeat(payload, count)
    elif sweep_start and sweep_end and sweep_end >= sweep_start:
        if not payload:
            payload = random_text(sweep_start)
        provider = payload_provider.Sweep(payload, sweep_start, sweep_end)
    options = ()
    if df:
        options = network.Socket.DONT_FRAGMENT

    # Fix to allow for pythonping multithreaded usage;
    # no need to protect this loop as no one will ever surpass 0xFFFF amount of threads
    while True:
        # seed_id needs to be less than or equal to 65535 (as original code was seed_id = getpid() & 0xFFFF)
        seed_id = randint(0x1, 0xFFFF)
        if seed_id not in SEED_IDs:
            SEED_IDs.append(seed_id)
            break

    comm = executor.Communicator(target, provider, timeout, socket_options=options, verbose=verbose, output=out,
                                 seed_id=seed_id)
    comm.run(match_payloads=match)

    SEED_IDs.remove(seed_id)

    return comm.responses



import unittest
from pythonping import ping


class PingCase(unittest.TestCase):
    """Tests for actual ping against localhost"""

    def test_ping_execution(self):
        """Verifies that random_text generates text of correct size"""
        # NOTE, this may be considered an e2e test
        self.assertEqual(len(ping('10.127.0.1', count=4, size=10)), 4,
                         'Sent 4 pings to localhost, but not received 4 responses')

        self.assertEqual(ping('8.8.8.8', count=4, size=992).success(), True,
                         'Sent 4 large pings to google DNS A with payload match off, received all replies')

        self.assertEqual(ping('8.8.8.8', count=4, size=992, match=True).success(), False,
                         'Sent 4 large pings to google DNS A with payload match on,'
                         + 'expected all to fail since they truncate large payloads')
